#include <stdio.h>
#include "common.h"
#include "spaceFillingCurve.h"
#include "radixSortModu.h"
#include "radixSortModu1.h"
#include "gpc.h"
 
extern void scan(int *g_data,int length,int size);
void printf(int numOfLayers,int * layer,int * prefix,int *rects_start,int *rects_end,int *rect1_rtree,int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int numOfNodes,int numOfRects,bounding_box *zOrder);
void printfSearch(rtree_search * output,int num);
void printCount(int * count,int num);
void printNode(rtree_node * nodes,int num);
void printCPUStruct(int numOfLayer,int * prefix,int *layer,int numOfNodes,int numOfRects);
void printBound(int * bounds_cpu,int numOfLayers);

void checkRtree(bounding_box * zOrder,int num);
void checkRtreeNode(int numOfNodes,int numOfRects,int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int * rect4_rtree,int * rects_start,int * rects_end);

__global__ void Construction(int offset1,int num1,int offset2,int num2,int * rects_start,int * rects_end,int *rect1_rtree,int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int *bound,int index);
__device__ void ResetMBR(int * rect1_rtree, int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int location);
__device__ void Merge(int * rect1_rtree, int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int location,int target1,int target2,int target3,int target4);
__device__ int Overlap(int rect1_target,int rect2_target,int rect3_target,int rect4_target,int rect1_source,int rect2_source,int rect3_source,int rect4_source);

__global__ void SearchWithShMRoot(int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int *rect4_rtree,int num,int *rect1_query,int *rect2_query,int *rect3_query,int *rect4_query,int bound,rtree_search * out,int numOfQuerys,int *count);
__global__ void SearchWithShMNonRoot(int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int *rect4_rtree,int *rect1_query,int *rect2_query,int *rect3_query,int * rect4_query,int multiple,int offset,int *rects_start,int *rects_end,int bound,rtree_search * in,rtree_search * out,rtree_node * nodes,int numOfQuerys,int *count,int * count_node);
__global__ void SearchWithoutShMNonRoot(int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int *rect4_rtree,int *rect1_query,int *rect2_query,int *rect3_query,int *rect4_query,int *rects_start,int *rects_end,int bound,rtree_search * in,rtree_search * out,int numOfQuerys,int *count);

__global__ void MemoryCopy(int * count,rtree_search * old_search,rtree_search * new_search,int numOfQuerys,int bound);
__global__ void aggregate(rtree_search *querys,int num,rtree_node *nodes,int offset);
__global__ void deduction(rtree_node *nodes,int count_node);
//__global__ void Execute(rtree_search *output,int offset,int * num_base,int * prefix_base,gpc_vertex * vertice_base,int *num_overlay,int * prefix_overlay, gpc_vertex * vertice_overlay,int num,gpc_polygon * result);

__global__ void reductionPerBlock(int *arr_old,int num,int * arr_new);
__global__ void reductionBetweenBlocks(int *array,int num);
int reduce(int *arr_old,int num);

void printf(int numOfLayers,int * layer,int * prefix,int *rects_start,int *rects_end,int *rect1_rtree,int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int numOfNodes,int numOfRects,bounding_box *zOrder){
	int i,j;
	int *rects_start_cpu;
	int *rects_end_cpu;
	int *rect1_rtree_cpu;
	int *rect2_rtree_cpu;
	int *rect3_rtree_cpu;
	int *rect4_rtree_cpu;
	bounding_box *zOrder_cpu;
	
	MALLOC(rects_start_cpu, numOfNodes * sizeof(int),"rect start creation", int);
	MALLOC(rects_end_cpu, numOfNodes * sizeof(int),"rect end creation", int);
	MALLOC(rect1_rtree_cpu, numOfRects * sizeof(int),"rect 1 creation", int);
	MALLOC(rect2_rtree_cpu, numOfRects * sizeof(int),"rect 2 creation", int);
	MALLOC(rect3_rtree_cpu, numOfRects * sizeof(int),"rect 3 creation", int);
	MALLOC(rect4_rtree_cpu, numOfRects * sizeof(int),"rect 4 creation", int);
	MALLOC(zOrder_cpu, (numOfRects-numOfNodes) * sizeof(bounding_box),"zOrder creation", bounding_box);

   	cudaMemcpy(rects_start_cpu,rects_start,sizeof(int)*numOfNodes,cudaMemcpyDeviceToHost);	
    cudaMemcpy(rects_end_cpu,rects_end,sizeof(int)*numOfNodes,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect1_rtree_cpu,rect1_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect2_rtree_cpu,rect2_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect3_rtree_cpu,rect3_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect4_rtree_cpu,rect4_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(zOrder_cpu,zOrder,sizeof(bounding_box)*(numOfRects-numOfNodes),cudaMemcpyDeviceToHost);	

	for(i=0;i<(numOfRects-numOfNodes);i++){
		//printf("zOrder %d:(%lu,%d)\n",i,zOrder_cpu[i].zOrderValue,zOrder_cpu[i].index);
	}
	
	for(i=numOfNodes;i<numOfRects;i++){
		if((rect1_rtree_cpu[i]<0)||(rect2_rtree_cpu[i]<0)||(rect3_rtree_cpu[i]<0)||(rect4_rtree_cpu[i]<0)){
			//printf("BB %d:(%d,%d,%d,%d)\n",(i-numOfNodes),rect1_rtree_cpu[i],rect2_rtree_cpu[i],rect3_rtree_cpu[i],rect4_rtree_cpu[i]);
		}
	}
	
	for(i=numOfLayers;i>=0;i--){
		printf("****Layer %d has %d elements*****\n",(numOfLayers-i),layer[i]);
		for(j=0;j<layer[i];j++){
			//printf("Node %d (%d,%d,%d,%d) (%d,%d)\n",prefix[i]+j,rect1_rtree_cpu[prefix[i]+j],rect2_rtree_cpu[prefix[i]+j],rect3_rtree_cpu[prefix[i]+j],rect4_rtree_cpu[prefix[i]+j],rects_start_cpu[prefix[i]+j],rects_end_cpu[prefix[i]+j]);
			/*if((rect1_rtree_cpu[prefix[i]+j]<0)||(rect2_rtree_cpu[prefix[i]+j]<0)||(rect3_rtree_cpu[prefix[i]+j]<0)||(rect4_rtree_cpu[prefix[i]+j]<0)){
				printf("Node %d is wrong\n", prefix[i]+j);
     			//exit(-1);
     			break;
			}*/
		}
		printf("\n");
	}
}

void printfSearch(rtree_search * output,int num){
	int i;
	rtree_search * output_cpu;
	printf("#############Search structure################\n");
	MALLOC(output_cpu, num * sizeof(rtree_search),"output_cpu", rtree_search);
	cudaMemcpy(output_cpu,output,sizeof(rtree_search)*num,cudaMemcpyDeviceToHost);	
	printf("num=%d\n",num);
	for(i=0;i<num;i++){	
		printf("(%d,%d)\n",output_cpu[i].queryid,output_cpu[i].nodeid);		
	}
	free(output_cpu);
}

void printCount(int * count,int num){
	int i;
	int * count_cpu;
	printf("#############Print Count################\n");
	printf("num=%d\n",num);
	MALLOC(count_cpu, num * sizeof(int),"count_cpu", int);
	cudaMemcpy(count_cpu,count,sizeof(int)*num,cudaMemcpyDeviceToHost);	
	for(i=0;i<num;i++){	
		printf("(%d)\n",count_cpu[i]);		
	}
	free(count_cpu);
}

void printNode(rtree_node * nodes,int num){
	int i;
	rtree_node * nodes_cpu;
	printf("#############Print Node################\n");
	MALLOC(nodes_cpu, num * sizeof(rtree_node),"nodes_cpu", rtree_node);
	cudaMemcpy(nodes_cpu,nodes,sizeof(rtree_node)*num,cudaMemcpyDeviceToHost);	
	printf("num=%d\n",num);
	for(i=0;i<num;i++){	
		printf("(%d,%d,%d)\n",nodes_cpu[i].start,nodes_cpu[i].end,nodes_cpu[i].flag);		
	}
	free(nodes_cpu);
}

void printCPUStruct(int numOfLayer,int * prefix,int *layer,int numOfNodes,int numOfRects){
	int i;
	printf("The num of layers=%d\n",(numOfLayer+1));
	printf("The num of nodes=%d\n",numOfNodes);
	printf("The num of rects=%d\n",numOfRects); 
	for(i=0;i<=numOfLayer;i++){
		printf("The %d element in the layer array has %d rects\n",i, layer[i]);
		printf("The %d element in the prefix array has %d rects ahead of it.\n",i, prefix[i]);
	}
}

void printBound(int * bounds_cpu,int numOfLayers){
	int i;
	printf("#############Bound################\n");
	for(i=numOfLayers;i>=0;i--){
		printf("layer %d's bound is %d\n",i,bounds_cpu[i]);
	}
}

void checkRtree(bounding_box * zOrder,int num){
	cudaError_t err;
	bounding_box *zOrder_cpu;
	int i;
	MALLOC(zOrder_cpu, num * sizeof(bounding_box),"zOrder creation", bounding_box);
	err=cudaMemcpy(zOrder_cpu,zOrder,sizeof(bounding_box)*num,cudaMemcpyDeviceToHost);	
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	exit(-1);
	}
	for(i=0;i<num;i++){
		//printf("(%lld,%d)\n",zOrder_cpu[i].zOrderValue,zOrder_cpu[i].index);
		if((zOrder_cpu[i].index>=num)||(zOrder_cpu[i].index<0)){
			printf("%d zOrder is wrong\n", i);
     		exit(-1);
		}
	}
}

void checkRtreeNode(int numOfNodes,int numOfRects,int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int * rect4_rtree,int * rects_start,int * rects_end){
	int i,j;
	int *rects_start_cpu;
	int *rects_end_cpu;
	int *rect1_rtree_cpu;
	int *rect2_rtree_cpu;
	int *rect3_rtree_cpu;
	int *rect4_rtree_cpu;
	
	MALLOC(rects_start_cpu, numOfNodes * sizeof(int),"rect start creation", int);
	MALLOC(rects_end_cpu, numOfNodes * sizeof(int),"rect end creation", int);
	MALLOC(rect1_rtree_cpu, numOfRects * sizeof(int),"rect 1 creation", int);
	MALLOC(rect2_rtree_cpu, numOfRects * sizeof(int),"rect 2 creation", int);
	MALLOC(rect3_rtree_cpu, numOfRects * sizeof(int),"rect 3 creation", int);
	MALLOC(rect4_rtree_cpu, numOfRects * sizeof(int),"rect 4 creation", int);

   	cudaMemcpy(rects_start_cpu,rects_start,sizeof(int)*numOfNodes,cudaMemcpyDeviceToHost);	
    cudaMemcpy(rects_end_cpu,rects_end,sizeof(int)*numOfNodes,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect1_rtree_cpu,rect1_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect2_rtree_cpu,rect2_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect3_rtree_cpu,rect3_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	cudaMemcpy(rect4_rtree_cpu,rect4_rtree,sizeof(int)*numOfRects,cudaMemcpyDeviceToHost);	
	/*for(i=numOfNodes;i<numOfRects;i++){
		if((rect1_rtree_cpu[i]<0)||(rect2_rtree_cpu[i]<0)||(rect3_rtree_cpu[i]<0)||(rect4_rtree_cpu[i]<0)){
			printf("Rectagle(%d,%d,%d,%d) %d is wrong\n",rect1_rtree_cpu[i],rect2_rtree_cpu[i],rect3_rtree_cpu[i],rect4_rtree_cpu[i],(i-numOfNodes));
     		exit(-1);
		}
	}
	for(i=0;i<numOfNodes;i++){
		if((rect1_rtree_cpu[i]<0)||(rect2_rtree_cpu[i]<0)||(rect3_rtree_cpu[i]<0)||(rect4_rtree_cpu[i]<0)){
			printf("Node(%d,%d,%d,%d) %d is wrong\n",rect1_rtree_cpu[i],rect2_rtree_cpu[i],rect3_rtree_cpu[i],rect4_rtree_cpu[i],i);
     		exit(-1);
		}
	}*/
}
/* Implement the bottom-up R-tree construction 
offset1,num1: offset and num in the current level.
offset2,num2: offset and num in the child level.
*/

__global__ void Construction(int offset1,int num1,int offset2,int num2,int * rects_start,int * rects_end,int *rect1_rtree,int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int *bound,int index){
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int thid = tx+bx*blockDim.x;
	long long int temp;
	int start;
	int end;
	int location;
	int r1,r2,r3,r4;
	if(thid<num1){
		temp=num2;
		temp=temp*thid;
		start=(int)(temp/num1);
		temp=num2;
		temp=temp*(thid+1);
		end=(int)(temp/num1-1);
		location=thid+offset1;
		rects_start[location]=start+offset2;
		rects_end[location]=end+offset2;
		ResetMBR(rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,location);
		for(int i=(start+offset2);i<=(end+offset2);i++){
 			r1=rect1_rtree[i];
 			r2=rect2_rtree[i];
 			r3=rect3_rtree[i];
 			r4=rect4_rtree[i]; 			
 			Merge(rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,location,rect1_rtree[i],rect2_rtree[i],rect3_rtree[i],rect4_rtree[i]);
			if(rect1_rtree[i]<0){
				//printf("(%d,%d)--(%d,%d,%d,%d)--(%d,%d,%d,%d)\n",(start+offset2),(end+offset2),rect1_rtree[i],rect2_rtree[i],rect3_rtree[i],rect4_rtree[i],r1,r2,r3,r4);
				break;				
			}
		}
	}
	if(thid==(num1-1)){
		bound[index]=end-start+1;
	}
}

/*__global__ void Construction(bounding_box * pointers,int num,rtree_node * nodes,int * root){
	int i,j,k;
	int numOfNodes;
	int numOfGroup;
	int numOfChildren;
	int index,childrenIndex;
	int begin,end;
	int tx=threadIdx.x;
	int bx=blockDim.x;
	long long int temp;
	index=0;
	numOfNodes=num;
	numOfGroup=numOfNodes/ m;	
	for(i=tx;i<numOfGroup;i=i+bx){
		temp=numOfNodes;
		temp=temp*i;
		begin=(int)(temp/numOfGroup);
		if(begin<0){
			childrenIndex=i;
		}
		temp=numOfNodes;
		temp=temp*(i+1);
		end=(int)(temp/numOfGroup-1);
		nodes[i].flag=LEAF;
		nodes[i].num=(end-begin+1);
		for(j=begin;j<=end;j++){
			nodes[i].data[j-begin]=pointers[j].data;
			nodes[i].mbr[j-begin].rect[0]=pointers[j].rect[0];
			nodes[i].mbr[j-begin].rect[1]=pointers[j].rect[1];
			nodes[i].mbr[j-begin].rect[2]=pointers[j].rect[2];
			nodes[i].mbr[j-begin].rect[3]=pointers[j].rect[3];
			nodes[i].children[j-begin]=-1;
		}
	}
	index=index+numOfGroup;
	numOfChildren=numOfGroup;
	childrenIndex=0;
	__syncthreads();
	while(numOfGroup>1){
		numOfNodes=numOfGroup;
		numOfGroup=numOfNodes/ m;
		for(i=index+tx;i<(numOfGroup+index);i=i+bx){
			temp=numOfNodes;
			temp=temp*(i-index);
			begin=(int)(temp/numOfGroup+childrenIndex);
			temp=numOfNodes;
			temp=temp*(i+1-index);
			end=(int)(temp/numOfGroup-1+childrenIndex);
			nodes[i].flag=NONLEAF;
			nodes[i].num=(end-begin+1);
			for(j=begin;j<=end;j++){
				nodes[i].children[j-begin]=j;
				nodes[i].data[j-begin]=-1;
				ResetMBR(&(nodes[i].mbr[j-begin]));
				for(k=0;k<nodes[j].num;k++){
					Merge(nodes[i].mbr[j-begin],nodes[j].mbr[k],&(nodes[i].mbr[j-begin]));
				}
			}
		}
		__syncthreads();
		index=index+numOfGroup;
		childrenIndex=childrenIndex+numOfChildren;
		numOfChildren=numOfGroup;
	}
	(*root)=index-numOfGroup;
}*/

__device__ void ResetMBR(int * rect1_rtree, int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int location){
	rect1_rtree[location]=INFINITE;
	rect2_rtree[location]=INFINITE;
	rect3_rtree[location]=NINFINITE;
	rect4_rtree[location]=NINFINITE;
}

__device__ void Merge(int * rect1_rtree, int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int location,int target1,int target2,int target3,int target4){
	if(rect1_rtree[location]>target1){
		rect1_rtree[location]=target1;
	}
	if(rect2_rtree[location]>target2){
		rect2_rtree[location]=target2;
	}
	if(rect3_rtree[location]<target3){
		rect3_rtree[location]=target3;
	}
	if(rect4_rtree[location]<target4){
		rect4_rtree[location]=target4;
	}	
}

__device__ int Overlap(int rect1_target,int rect2_target,int rect3_target,int rect4_target,int rect1_source,int rect2_source,int rect3_source,int rect4_source){

	int rect1_temp;
	int rect2_temp;
	int rect3_temp;
	int rect4_temp;
	if(rect1_source>rect1_target){
		rect1_temp=rect1_target;
	}
	else{
		rect1_temp=rect1_source;
	}
	if(rect2_source>rect2_target){
		rect2_temp=rect2_target;
	}
	else{
		rect2_temp=rect2_source;
	}
	if(rect3_source<rect3_target){
		rect3_temp=rect3_target;
	}
	else{
		rect3_temp=rect3_source;
	}
	if(rect4_source<rect4_target){
		rect4_temp=rect4_target;
	}
	else{
		rect4_temp=rect4_source;
	}
	if((rect3_target-rect1_target+rect3_source-rect1_source)<(rect3_temp-rect1_temp)){
		return 0;
	}
	if((rect4_target-rect2_target+rect4_source-rect2_source)<(rect4_temp-rect2_temp)){
		return 0;
	}
	return 1;
}

__global__ void SearchWithShMRoot(int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int *rect4_rtree,int num,int *rect1_query,int *rect2_query,int *rect3_query,int *rect4_query,int bound,rtree_search * out,int numOfQuerys,int *count){
	int blockid=blockIdx.x;
	int threadid=threadIdx.x;
	int thid=blockid*blockDim.x+threadid;
	int sum=0;
	int i;
	extern __shared__ int temp_1[]; //allocated on invocation 
	for(i=threadid;i<num;i=i+blockDim.x){
		temp_1[i]=rect1_rtree[i+1];
		temp_1[i+num]=rect2_rtree[i+1];
		temp_1[i+num+num]=rect3_rtree[i+1];
		temp_1[i+num+num+num]=rect4_rtree[i+1];
	}
	__syncthreads(); 
	if(thid<numOfQuerys){
		for(i=0;i<num;i++){
			if(Overlap(temp_1[i],temp_1[i+num],temp_1[i+num+num],temp_1[i+num+num+num],rect1_query[thid],rect2_query[thid],rect3_query[thid],rect4_query[thid])){
				out[thid*bound+sum].queryid=thid;
				out[thid*bound+sum].nodeid=(i+1);
				sum++;
			}
		}
		count[thid]=sum;
	}
}

__global__ void SearchWithShMNonRoot(int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int *rect4_rtree,int *rect1_query,int *rect2_query,int *rect3_query,int * rect4_query,int multiple,int offset,int *rects_start,int *rects_end,int bound,rtree_search * in,rtree_search * out,rtree_node * nodes,int numOfQuerys,int *count,int * count_node){
	int blockid=blockIdx.x;
	int threadid=threadIdx.x;
	//int thid=blockid*blockDim.x+threadid;
	int index=blockid / multiple;
	int remaining=blockid % multiple;
	int sum=0;
	int start;
	int end;
	int start1;
	int end1;
	//int totalQuerys;
	int num;
	int temp1;
	int temp2;
	int queryid;
	int i;
	extern __shared__ int temp_2[]; //allocated on invocation 
	if(nodes[index].flag==1){
		start=nodes[index].start;
		end=nodes[index].end;
		//totalQuerys=end-start+1;
		start1=start+remaining*(blockDim.x);
		end1=start+(remaining+1)*(blockDim.x)-1;
		if(start1<=end){ 
			if(end1>end){
				end1=end;
			}			
			temp1=rects_start[index+offset];
			temp2=rects_end[index+offset];
			num=(temp2-temp1+1);
			for(i=threadid;i<num;i=i+blockDim.x){
				temp_2[i]=rect1_rtree[i+temp1];
				temp_2[i+num]=rect2_rtree[i+temp1];
				temp_2[i+num+num]=rect3_rtree[i+temp1];
				temp_2[i+num+num+num]=rect4_rtree[i+temp1];
			}
			__syncthreads(); 
			if(threadid+start1<=end1){
				queryid=in[threadid+start1].queryid;
				for(i=0;i<num;i++){
					if(Overlap(temp_2[i],temp_2[i+num],temp_2[i+num+num],temp_2[i+num+num+num],rect1_query[queryid],rect2_query[queryid],rect3_query[queryid],rect4_query[queryid])){
						out[(threadid+start1)*bound+sum].queryid=queryid;
						out[(threadid+start1)*bound+sum].nodeid=temp1+i;
						sum++;
					}
				}
				if(index==0){
					count[start1+threadid]=sum;
				}
				else{
					//count[start1+threadid+count_node[index-1]]=sum;
					count[start1+threadid]=sum;
				}
			}	
		}	
	}
}

__global__ void SearchWithoutShMNonRoot(int * rect1_rtree,int * rect2_rtree,int * rect3_rtree,int *rect4_rtree,int *rect1_query,int *rect2_query,int *rect3_query,int *rect4_query,int *rects_start,int *rects_end,int bound,rtree_search * in,rtree_search * out,int numOfQuerys,int *count){
	int blockid=blockIdx.x;
	int threadid=threadIdx.x;
	int thid=blockid*blockDim.x+threadid;
	int queryid,nodeid;
	int sum=0;
	int start;
	int end;
	int i;
	if(thid<numOfQuerys){
		queryid=in[thid].queryid;
		nodeid=in[thid].nodeid;
		start=rects_start[nodeid];
		end=rects_end[nodeid];
		for(i=start;i<end;i++){
			if(Overlap(rect1_rtree[i],rect2_rtree[i],rect3_rtree[i],rect4_rtree[i],rect1_query[queryid],rect2_query[queryid],rect3_query[queryid],rect4_query[queryid])){
				out[thid*bound+sum].queryid=queryid;
				out[thid*bound+sum].nodeid=i;
				sum++;
			}
		}
		count[thid]=sum;
	}
}


__global__ void MemoryCopy (int * count,rtree_search * old_search,rtree_search * new_search,int numOfQuerys,int bound){
	int blockid=blockIdx.x;
	int threadid=threadIdx.x;
	int thid=blockid*blockDim.x+threadid;
	int i;
	if(thid<numOfQuerys){
		if(thid==0){
			for(i=0;i<count[thid];i++){
				new_search[i].queryid=old_search[i].queryid;
				new_search[i].nodeid=old_search[i].nodeid;
			}	
		}
		else{
			for(i=0;i<(count[thid]-count[thid-1]);i++){
				new_search[count[thid-1]+i]=old_search[thid*bound+i];
			}	
		}
	}
}

__global__ void aggregate(rtree_search *querys,int num,rtree_node *nodes,int offset){
	int blockid=blockIdx.x;
	int threadid=threadIdx.x;
	int id=blockid*blockDim.x+threadid;
	if(id<=num-1){
		if(id==0){
			nodes[querys[0].nodeid-offset].start=0;
			nodes[querys[0].nodeid-offset].flag=1;
		}
		else if(id==(num-1)){
			nodes[querys[id].nodeid-offset].end=num-1;
		}
		else {
			if(querys[id].nodeid!=querys[id+1].nodeid){
				nodes[querys[id].nodeid-offset].end=id;
				nodes[querys[id+1].nodeid-offset].start=id+1;
				nodes[querys[id+1].nodeid-offset].flag=1;
			}
		}
	}
}

__global__ void deduction(rtree_node *nodes,int *count_node,int num){
	int blockid=blockIdx.x;
	int threadid=threadIdx.x;
	int id=blockid*blockDim.x+threadid;
	if(id<num){
		if(nodes[id].flag){
			count_node[id]=nodes[id].end-nodes[id].start+1;
		}
		else{
			count_node[id]=0;
		}
	}
}

/*__global__ void Execute(rtree_search *output,int offset,int * num_base,int * prefix_base,gpc_vertex * vertice_base,int *num_overlay,int * prefix_overlay, gpc_vertex * vertice_overlay,int num,gpc_polygon * result){
	int thid=blockIdx.x*blockDim.x+threadIdx.x;
	gpc_polygon base;
	gpc_vertex_list list_base;
	gpc_polygon overlay;
	gpc_vertex_list list_overlay;
	if(thid<num){
		list_base.num_vertices=num_base[output[thid].nodeid-offset];
		list_base.vertex=&(vertice_base[prefix_base[output[thid].nodeid-offset]]);
		base.num_contours=1;
		base.contour=&list_base;
		
		list_overlay.num_vertices=num_overlay[output[thid].queryid];
		list_overlay.vertex=&(vertice_overlay[prefix_overlay[output[thid].queryid]]);
		overlay.num_contours=1;
		overlay.contour=&list_overlay;
		gpc_polygon_clip(GPC_INT, &base, &overlay,result);
	}
}*/

__global__ void reductionPerBlock(int *arr_old,int num,int * arr_new){
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int id=bx*blockDim.x+tx;
	int * arr=(arr_old)+bx*blockDim.x;
	int size=blockDim.x;
	int round=1;
	while(round<size){
		if((tx%round==0)&&((id+round)<num)){
			//arr[tx]=arr[tx+round]+arr[tx];
			if(arr[tx]<arr[tx+round]){
				arr[tx]=arr[tx+round];
			}
		}
		round=round<<1;
		__syncthreads();
	}
	if(tx==0){
		arr_new[bx]=arr[0];
	}
}

__global__ void reductionBetweenBlocks(int *array,int num){
	int tx=threadIdx.x;
	int round=1;
	while(round<num){
		if((tx % round==0)&&((tx+round)<num)){
			//array[tx]=array[tx+round]+array[tx];
			if(array[tx]<array[tx+round]){
				array[tx]=array[tx+round];
			}
		}
		round=round<<1;		
		__syncthreads();
	}
}

int reduce(int *arr_old,int num){
	int numOfBlocks=(num-1)/512+1;
	int result;
	int *arr_new;
	if(numOfBlocks==1){
		dim3 dimBlock(512,1);
    	dim3 dimGrid(1,1);
    	reductionBetweenBlocks<<<dimGrid,dimBlock>>>(arr_old,num);
		(cudaMemcpy(&result,arr_old, sizeof(int)*1, cudaMemcpyDeviceToHost));
	}
	else{
		dim3 dimBlock1(512,1);
    	dim3 dimGrid1(numOfBlocks,1);
    	(cudaMalloc( (void**) &arr_new, sizeof(int)*numOfBlocks));
    	reductionPerBlock<<<dimGrid1,dimBlock1>>>(arr_old,num,arr_new);
		dim3 dimBlock2(512,1);
    	dim3 dimGrid2(1,1);
		reductionBetweenBlocks<<<dimGrid2,dimBlock2>>>(arr_new,numOfBlocks);
		(cudaMemcpy(&result,arr_new, sizeof(int)*1, cudaMemcpyDeviceToHost));
		cudaFree(arr_new);
	}
	return result;
}

int main(int argc , char **argv){

  	int *rect1;
  	int *rect2;
  	int *rect3;
  	int *rect4;
  	int *data;
  	int num_contours;
  	int *rect1_gpu;
	int *rect2_gpu;
	int *rect3_gpu;
	int *rect4_gpu;
	int *data_gpu;
	
  	long long int *x1;
  	long long int *x2;
  	long long int *x3;
  	long long int *x4;
  	long long int *y1;
  	long long int *y2;
  	long long int *y3;
  	long long int *y4;
	bounding_box *zOrder;
  	int i;
  	int m;
  	int M;
	
	h_state arr[3][6]={{BH, TH,   TH, BH,   NH, NH}, {NH, NH,   NH, NH,   TH, TH}, {NH, NH,   NH, NH,   BH, BH}};
	cudaMemcpyToSymbol("next_h_state", arr, 18*sizeof(h_state), 0, cudaMemcpyHostToDevice);
	
	size_t f, t;
	cudaMemGetInfo(&f, &t);
	printf("Before running program available=%lu,total=%lu,consumed=%lu\n",(unsigned long int)f,(unsigned long int)t,(unsigned long int)(t-f));

	int *buf_d = NULL;
	cudaError_t err;
	err = cudaMalloc((void **)&buf_d, 1000000000*sizeof(int));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	cudaMemGetInfo(&f, &t);
	printf("After allocation, program available=%lu,total=%lu,consumed=%lu\n",(unsigned long int)f,(unsigned long int)t,(unsigned long int)(t-f));

	char * file1=argv[1];
	char * file2=argv[2];
	m=atoi(argv[3]);
	M=atoi(argv[4]);
	
  	//fp=fopen("data/Rect1000000.txt","r"); 
  	/*
  	FILE *fp; 
  	fp=fopen("data/sample1.txt","r");
  	fscanf(fp, "%d", &num_contours);
    rect1=(int *)malloc(num_contours*sizeof(int));
    rect2=(int *)malloc(num_contours*sizeof(int));
    rect3=(int *)malloc(num_contours*sizeof(int));
    rect4=(int *)malloc(num_contours*sizeof(int));
    data=(int *)malloc(num_contours*sizeof(int));

	int minX,maxX,minY,maxY;
	minX=10000000;
	maxX=-10000000;
	minY=10000000;
	maxY=-10000000;
  	for (i= 0; i < num_contours; i++) 
  	{
    	fscanf(fp, "%d %d", &(rect1[i]),&(rect2[i]));
    	fscanf(fp, "%d %d", &(rect3[i]),&(rect4[i]));
    	fscanf(fp, "%d", &(data[i]));
    	//key++;
		if(rect1[i]<minX){
			minX=rect1[i];
		}
		if(rect3[i]>maxX){
			maxX=rect3[i];
		}
		if(rect2[i]<minY){
			minY=rect2[i];
		}
		if(rect4[i]>maxY){
			maxY=rect4[i];
		}
    }
	*/
	
	
	int numOfQuerys;
	int *rect1_query_cpu;
	int *rect2_query_cpu;
	int *rect3_query_cpu;
	int *rect4_query_cpu;
	int *rect1_query;
	int *rect2_query;
	int *rect3_query;
	int *rect4_query;
	
	/*
	FILE *fp_query;
	fp_query=fopen("data/query.txt","r");	
  	fscanf(fp_query, "%d", &numOfQuerys);
	//MALLOC(rect1_query_cpu, numOfQuerys * sizeof(int),"rect1_query creation", int);
	//MALLOC(rect2_query_cpu, numOfQuerys * sizeof(int),"rect2_query creation", int);
	//MALLOC(rect3_query_cpu, numOfQuerys * sizeof(int),"rect3_query creation", int);
	//MALLOC(rect4_query_cpu, numOfQuerys * sizeof(int),"rect4_query creation", int);
	(cudaMalloc( (void**) &rect1_query, sizeof(int)*numOfQuerys));
	(cudaMalloc( (void**) &rect2_query, sizeof(int)*numOfQuerys));
	(cudaMalloc( (void**) &rect3_query, sizeof(int)*numOfQuerys));
	(cudaMalloc( (void**) &rect4_query, sizeof(int)*numOfQuerys));
	for(i=0;i<numOfQuerys;i++){
	    fscanf(fp_query, "%d %d %d %d",(rect1_query_cpu+i),(rect2_query_cpu+i),(rect3_query_cpu+i),(rect4_query_cpu+i));
	}*/

	int *num_base;
	long int *prefix_base;
	gpc_vertex * vertex_base;
	//int numOfVertices_base;
	int *num_overlay;
	long int *prefix_overlay;
	gpc_vertex *vertex_overlay;
	//int numOfVertices_overlay;
	//readData(&rect1,&rect2,&rect3,&rect4,&num_contours,&rect1_query,&rect2_query,&rect3_query,&rect4_query,&numOfQuerys,&num_base,&prefix_base,&vertice_base,&numOfVertices_base,&num_overlay,&prefix_overlay,&vertice_overlay,&numOfVertices_overlay);
	int minX,minY;
	minX=10000000;
	minY=10000000;	

	readData(file1,file2, &rect1,&rect2,&rect3,&rect4,&num_contours,&rect1_query_cpu,&rect2_query_cpu,&rect3_query_cpu,&rect4_query_cpu,&numOfQuerys,&num_base,&prefix_base,&vertex_base,&num_overlay,&prefix_overlay,&vertex_overlay,&minX,&minY);

	//rect1_query_cpu=(int *)malloc(numOfQuerys*sizeof(int));
  	//rect2_query_cpu=(int *)malloc(numOfQuerys*sizeof(int));
  	//rect3_query_cpu=(int *)malloc(numOfQuerys*sizeof(int));
  	//rect4_query_cpu=(int *)malloc(numOfQuerys*sizeof(int));
   	//(cudaMalloc( (void**) &rect1_query, sizeof(int)*numOfQuerys));
	//(cudaMalloc( (void**) &rect2_query, sizeof(int)*numOfQuerys));
	//(cudaMalloc( (void**) &rect3_query, sizeof(int)*numOfQuerys));
	//(cudaMalloc( (void**) &rect4_query, sizeof(int)*numOfQuerys));
   	//(cudaMemcpy( rect1_query, rect1_query_cpu, sizeof(int)*numOfQuerys, cudaMemcpyHostToDevice));
   	//(cudaMemcpy( rect2_query, rect2_query_cpu, sizeof(int)*numOfQuerys, cudaMemcpyHostToDevice));
   	//(cudaMemcpy( rect3_query, rect3_query_cpu, sizeof(int)*numOfQuerys, cudaMemcpyHostToDevice));
   	//(cudaMemcpy( rect4_query, rect4_query_cpu, sizeof(int)*numOfQuerys, cudaMemcpyHostToDevice));     
    
	cudaEvent_t start_event, stop_event;
    cudaEventCreate(&start_event) ;
    cudaEventCreate(&stop_event) ;
    cudaEventRecord(start_event, 0);

	err =(cudaMalloc( (void**) &rect1_gpu, sizeof(int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &rect2_gpu, sizeof(int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &rect3_gpu, sizeof(int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &rect4_gpu, sizeof(int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	//(cudaMalloc( (void**) &data_gpu, sizeof(int)*num_contours));
	err =(cudaMalloc( (void**) &x1, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &x2, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &x3, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &x4, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &y1, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &y2, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &y3, sizeof(long long int)*num_contours));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &y4, sizeof(long long int)*num_contours));
   	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
    int len=((num_contours-1)/512+1)*512;
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
	err =(cudaMalloc( (void**) &zOrder, sizeof(bounding_box)*len));   
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	} 
	err=(cudaMemcpy( rect1_gpu, rect1, sizeof(int)*num_contours, cudaMemcpyHostToDevice));
	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}	
	(cudaMemcpy( rect2_gpu, rect2, sizeof(int)*num_contours, cudaMemcpyHostToDevice));
	(cudaMemcpy( rect3_gpu, rect3, sizeof(int)*num_contours, cudaMemcpyHostToDevice));
	(cudaMemcpy( rect4_gpu, rect4, sizeof(int)*num_contours, cudaMemcpyHostToDevice));
	
	dim3 dimBlock(512,1);
    dim3 dimGrid((num_contours-1)/512+1,1);
	CalcuateZOrderValue<<<dimGrid, dimBlock>>>(rect1_gpu,rect2_gpu,rect3_gpu,rect4_gpu,x1,x2,x3,x4,y1,y2,y3,y4,zOrder,num_contours,minX,minY);
	
	//checkRtree(zOrder,num_contours);	

	/*
	bounding_box *zOrder_cpu;
	MALLOC(zOrder_cpu,num_contours * sizeof(bounding_box),"zOrder creation", bounding_box);
	cudaMemcpy(zOrder_cpu,zOrder,sizeof(bounding_box)*num_contours,cudaMemcpyDeviceToHost);	
	for(i=0;i<num_contours;i++){
		printf("zOrder %d:(%lu,%d)\n",i,zOrder_cpu[i].zOrderValue,zOrder_cpu[i].index);
	}
	*/

	sortGM(zOrder,num_contours);
	checkRtree(zOrder,num_contours);	

    int layer[30];
    int prefix[30];
    int numOfGroups=num_contours/m;
    int numOfLayers=0;
    int numOfNodes=0;
    int numOfRects=num_contours;
    while(numOfGroups>1){
    	layer[numOfLayers]=numOfGroups;
    	numOfGroups=numOfGroups/m;
    	numOfLayers++;
    }  
    layer[numOfLayers]=1;
    //numOfLayer++;
    prefix[numOfLayers]=0;
    for(i=numOfLayers-1;i>=0;i--){
    	prefix[i]=prefix[i+1]+layer[i+1];
    }
    numOfNodes=prefix[0]+layer[0];
    numOfRects=numOfRects+numOfNodes;
 	//printCPUStruct(numOfLayers,prefix,layer,numOfNodes,numOfRects);
    int *rects_start;
    int *rects_end;
   	err=(cudaMalloc( (void**) &rects_start, sizeof(int)*numOfNodes));
   	if( err != cudaSuccess)
	{
    	printf("CUDA error: %s\n", cudaGetErrorString(err));
     	return -1;
	}
   	(cudaMalloc( (void**) &rects_end, sizeof(int)*numOfNodes));
	int *rect1_rtree;
	int *rect2_rtree;
	int *rect3_rtree;
	int *rect4_rtree;
	int *bounds;
	
	(cudaMalloc( (void**) &rect1_rtree, sizeof(int)*numOfRects));
	//checkErrors("Memory check1");
	(cudaMalloc( (void**) &rect2_rtree, sizeof(int)*numOfRects));
	//checkErrors("Memory check2");
	(cudaMalloc( (void**) &rect3_rtree, sizeof(int)*numOfRects));
	//checkErrors("Memory check3");
	(cudaMalloc( (void**) &rect4_rtree, sizeof(int)*numOfRects));
	//checkErrors("Memory check4");
	(cudaMalloc( (void**) &bounds, sizeof(int)*(numOfLayers+1)));
	//checkErrors("Memory check5");
	moveData(zOrder,rect1_gpu,rect2_gpu,rect3_gpu,rect4_gpu,rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,(prefix[0]+layer[0]),num_contours);
	checkErrors("Memory check6");
	
	cudaMemGetInfo(&f, &t);
	//printf("After building the R-tree, available=%lu,total=%lu,consumed=%lu\n",(unsigned long int)f,(unsigned long int)t,(unsigned long int)(t-f));
	
	cudaFree(rect1_gpu);
	cudaFree(rect2_gpu);
	cudaFree(rect3_gpu);
	cudaFree(rect4_gpu);
	
	cudaFree(x1);
	cudaFree(x2);
	cudaFree(x3);
	cudaFree(x4);
	cudaFree(y1);
	cudaFree(y2);
	cudaFree(y3);
	cudaFree(y4);
	free(rect1);
	free(rect2);
	free(rect3);
	free(rect4);


	cudaMemGetInfo(&f, &t);
	
	printf("After freeing the memory, available=%lu,total=%lu,consumed=%lu\n",(unsigned long int)f,(unsigned long int)t,(unsigned long int)(t-f));
		
	dim3 dimBlockCons(512,1);
    dim3 dimGridCons((layer[0]-1)/512+1,1);
	Construction<<<dimGridCons,dimBlockCons>>>(prefix[0],layer[0],numOfNodes,num_contours,rects_start,rects_end,rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,bounds,0);
	for(i=1;i<=numOfLayers;i++){
	    dim3 dimGridCons1(layer[i]/512+1,1);
	    dim3 dimBlockCons1(512,1);
		Construction<<<dimGridCons1,dimBlockCons1>>>(prefix[i],layer[i],prefix[i-1],layer[i-1],rects_start,rects_end,rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,bounds,i);
	}
	printf(numOfLayers,layer,prefix,rects_start,rects_end,rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,numOfNodes,numOfRects,zOrder);
	cudaFree(zOrder);
	checkRtreeNode(numOfNodes,numOfRects,rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,rects_start,rects_end);

	
    /*int *rects_start_cpu;
    int *rects_end_cpu;
    MALLOC(rects_start_cpu, numOfNodes * sizeof(int),"rect start creation", int);
    MALLOC(rects_end_cpu, numOfNodes * sizeof(int),"rect end creation", int);
   	(cudaMemcpy( rects_start_cpu, rects_start_cpu, sizeof(int)*numOfNodes, cudaMemcpyDeviceToHost));
   	(cudaMemcpy( rects_end_cpu, rects_end_cpu, sizeof(int)*numOfNodes, cudaMemcpyDeviceToHost));*/
   	
   	/*
   	int * bounds_cpu;
   	MALLOC(bounds_cpu, (numOfLayers+1) * sizeof(int),"bounds array creation", int);
	(cudaMemcpy( bounds_cpu, bounds, sizeof(int)*(numOfLayers+1), cudaMemcpyDeviceToHost));
   	printBound(bounds_cpu,numOfLayers);

   	int numOfOutput;
   	rtree_search * input;
    rtree_search * output;
    rtree_search * output_after;
    int *count;
    rtree_node *nodes;
    int len1;
    int len2;
    int bound;
    int limit=0;
    int multiple;
    
	int counter=numOfLayers;
	(cudaMalloc( (void**) &output, sizeof(rtree_search)*numOfQuerys*layer[numOfLayers-1]));
	len2=((numOfQuerys-1)/512+1)*512;
	(cudaMalloc( (void**) &count, sizeof(int)*len2));
	dim3 dimBlockSearch1(512,1);
	dim3 dimGridSearch1((numOfQuerys-1)/512+1,1);
	SearchWithShMRoot<<<dimGridSearch1,dimBlockSearch1,4*layer[numOfLayers-1]*sizeof(int)>>>(rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,layer[numOfLayers-1],rect1_query,rect2_query,rect3_query,rect4_query,layer[numOfLayers-1],output,numOfQuerys,count);
	printf("*********************Search Algorithm 1***************************\n");
	printCount(count,numOfQuerys);
	scan(count,numOfQuerys,512);
	cudaMemcpy(&numOfOutput,&(count[numOfQuerys-1]),sizeof(int),cudaMemcpyDeviceToHost);	
	len1=((numOfOutput-1)/512+1)*512;
	cudaMalloc((void**)&output_after,sizeof(rtree_search)*len1);
	dim3 dimBlockMov1(512,1);
	dim3 dimGridMov1((numOfOutput-1)/512+1,1);
	MemoryCopy<<<dimGridMov1,dimBlockMov1>>>(count,output,output_after,numOfQuerys,layer[numOfLayers-1]);
	printfSearch(output_after,numOfOutput);
	counter--;
	int * count_node;
	int numOfMaxNode;
	int numOfInput;
	while((counter>=0)&&(limit<2)){
		sortGM1(output_after,numOfOutput);
		printfSearch(output_after,numOfOutput);
		cudaMalloc((void**)&nodes,sizeof(rtree_node)*layer[numOfLayers-1-limit]);
		len2=((layer[numOfLayers-1-limit]-1)/512+1)*512;
		cudaMalloc((void**)&count_node,sizeof(int)*len2);
		dim3 dimBlockAgg1(512,1);
		dim3 dimGridAgg1((numOfOutput-1)/512+1,1);
		aggregate<<<dimGridAgg1,dimBlockAgg1>>>(output_after,numOfOutput,nodes,prefix[numOfLayers-1-limit]);
		printNode(nodes,layer[numOfLayers-1-limit]);
		dim3 dimBlockDed1(512,1);
		dim3 dimGridDed1((layer[numOfLayers-1-limit]-1)/512+1,1);
		deduction<<<dimGridDed1,dimBlockDed1>>>(nodes,count_node,layer[numOfLayers-1-limit]);
		scan(count_node,layer[numOfLayers-1-limit],512);
		//reduce(count_node,layer[numOfLayers-1-limit]);
		cudaMemcpy(&numOfMaxNode,&(count_node[layer[numOfLayers-1-limit]-1]),sizeof(int),cudaMemcpyDeviceToHost);	
		multiple=(numOfMaxNode-1)/512+1;
		bound=bounds_cpu[numOfLayers-1-limit];
		input=output_after;
		cudaFree(output);
		cudaFree(count);
		len2=((numOfOutput-1)/512+1)*512;
		(cudaMalloc( (void**) &count, sizeof(int)*len2)); 
		(cudaMalloc( (void**) &output, sizeof(rtree_search)*numOfOutput*bound));
		dim3 dimBlockSearch2(512,1);
		dim3 dimGridSearch2(layer[numOfLayers-1-limit]*multiple,1);
		printf("*********************Search Algorithm 2***************************\n");
		SearchWithShMNonRoot<<<dimGridSearch2,dimBlockSearch2,4*bound*sizeof(int)>>>(rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,rect1_query,rect2_query,rect3_query,rect4_query,multiple,prefix[numOfLayers-1-limit],rects_start,rects_end,bound,input,output,nodes,numOfOutput,count,count_node);
		printf("bound=%d\n",bound);
		printfSearch(output,numOfOutput*bound);
		printCount(count,numOfOutput);
		cudaFree(input);
		cudaFree(count_node);
		cudaFree(nodes);
		//printfSearch(output,numOfOutput*bound);
		scan(count,numOfOutput,512);
		numOfInput=numOfOutput;
		cudaMemcpy(&numOfOutput,&(count[numOfOutput-1]),sizeof(int),cudaMemcpyDeviceToHost);	
		len1=((numOfOutput-1)/512+1)*512;
		cudaMalloc((void**)&output_after,sizeof(rtree_search)*len1);
		dim3 dimBlockMov2(512,1);
		dim3 dimGridMov2((numOfInput-1)/512+1,1);
		MemoryCopy<<<dimGridMov2,dimBlockMov2>>>(count,output,output_after,numOfInput,bound);
		printfSearch(output_after,numOfOutput);
		counter--;
		limit++;
	}
	while(counter>=0){
		input=output_after;
		scan(count,numOfOutput,512);
		//cudaMemcpy(&numOfOutput,&(count[numOfOutput-1]),sizeof(int),cudaMemcpyDeviceToHost);	
		//len1=((numOfOutput-1)/512+1)*512;
		cudaFree(output);
		cudaFree(count);
		len2=((numOfOutput-1)/512+1)*512;
		(cudaMalloc( (void**) &count, sizeof(int)*len2)); 
		(cudaMalloc( (void**) &output, sizeof(rtree_search)*numOfOutput*bound));
		dim3 dimBlockSearch3(512,1);
		dim3 dimGridSearch3(((numOfOutput-1)/512+1),1);
		printf("*********************Search Algorithm 3***************************\n");
		bound=bounds_cpu[numOfLayers-1-limit];
		printf("bound=%d\n",bound);
		SearchWithoutShMNonRoot<<<dimGridSearch3,dimBlockSearch3>>>(rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,rect1_query,rect2_query,rect3_query,rect4_query,rects_start,rects_end,bound,input,output,numOfOutput,count);
		printfSearch(output,numOfOutput*bound);
		printCount(count,numOfOutput);
		cudaFree(input);
		scan(count,numOfOutput,512);
		numOfInput=numOfOutput;
		cudaMemcpy(&numOfOutput,&(count[numOfOutput-1]),sizeof(int),cudaMemcpyDeviceToHost);	
		len1=((numOfOutput-1)/512+1)*512;
		cudaMalloc((void**)&output_after,sizeof(rtree_search)*len1);
		dim3 dimBlockMov3(512,1);
		dim3 dimGridMov3((numOfInput-1)/512+1,1);
		MemoryCopy<<<dimGridMov3,dimBlockMov3>>>(count,output,output_after,numOfInput,bound);
		printfSearch(output_after,numOfOutput);
		counter--;
		limit++;
	}
    cudaFree(count);
    cudaFree(output);
    printf("The total number of pairs is %d\n", numOfOutput);
    gpc_polygon ** result;
    cudaMalloc((void***)&result,sizeof(gpc_polygon *)*numOfOutput);
    dim3 dimBlockExecute(512,1);
	dim3 dimGridExecute((numOfOutput-1)/512+1,1);
	//__global__ void Execute(rtree_search *output,int offset,int * num_base,int * prefix_base,gpc_vertex * vertice_base,int *num_overlay,int * prefix_overlay, gpc_vertex * vertice_overlay,int num,gpc_polygon * result){
    Execute<<<dimGridExecute,dimBlockExecute>>>(output_after,numOfNodes,num_base,prefix_base,vertex_base,num_overlay,prefix_overlay,vertex_overlay,numOfOutput,result);
    */
    cudaEventRecord(stop_event, 0);
    cudaEventSynchronize(stop_event);	
    float time_kernel; 
    cudaEventElapsedTime(&time_kernel, start_event, stop_event);
	printf("Total time %f\n", time_kernel);
    return 0;
}
