#include "Sort.h"

__global__ void LocalSort(datatype *g_data,datatype *g_data_bak,int *histogram,int size,int round,int numOfBlocks){
	__shared__ key temp[512]; //allocated on invocation 
	__shared__ int temp1[512];
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	int i;
	int pout = 0, pin = 1;
	int out=1,in=0;
	int zeros;
	int offset;
	temp[tx].zOrderValue = g_data[thid].zOrderValue;
	//temp[tx].zOrderValue=g_data[thid].data;
	temp[tx].index=tx;
	for(i=0;i<BIT;i++){
		//Count the number of zeros
		temp1[pout*size+tx]=(1-((temp[in*size+tx].zOrderValue>>(i+round))&0x00000001));
		__syncthreads();
		for (offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp1[pout*size+tx] = temp1[pin*size+tx-offset]+temp1[pin*size+tx];
			}
			else {
				temp1[pout*size+tx] = temp1[pin*size+tx];
			}
			__syncthreads(); 
		}
		temp[out*size+tx].index=temp1[pout*size+tx]*(1-(temp[in*size+tx].zOrderValue>>(i+round))&0x00000001);
		zeros=temp1[pout*size+size-1];
		//missing
		__syncthreads();
		temp1[pout*size+tx]=((temp[in*size+tx].zOrderValue>>(i+round))&0x00000001);
		__syncthreads();
		for (offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp1[pout*size+tx] = temp1[pin*size+tx-offset]+temp1[pin*size+tx];
			}
			else {
				temp1[pout*size+tx] = temp1[pin*size+tx];
			}
			__syncthreads(); 
		}
		temp1[pin*size+tx]=(temp1[pout*size+tx]+zeros)*((temp[in*size+tx].zOrderValue>>(i+round))&0x00000001)+temp[out*size+tx].index;
		__syncthreads(); 
		temp[out*size+(temp1[pin*size+tx]-1)]=temp[in*size+tx];
		out = 1 - out; // swap double buffer indices 
		in = 1 - out;
		__syncthreads(); 
	}
	g_data_bak[thid]=g_data[(temp[in*size+tx].index)+base];  
	for(i=0;i<BYTES;i++){
		if(((temp[in*size+tx].zOrderValue>>(round))&0x0000000F)==i){
			temp1[pin*size+tx]=1;
		}
		else{
			temp1[pin*size+tx]=0;
		}
		__syncthreads();
		//g_data[thid]=temp[pin*size+txx]; 
		int offset=size;		
		while(offset>1){
			offset=(offset>>1);
			if(tx<offset){
				temp1[pin*size+tx]=temp1[pin*size+tx+offset]+temp1[pin*size+tx];
			}
			__syncthreads();
		}
		if(tx==0){
			histogram[i*numOfBlocks+bx]=temp1[pin*size+tx];
		}
		__syncthreads(); 
	}
}
__global__ void Reorder(datatype *g_data,datatype *g_data_out,int *histogram,int *position,int size,int round,int numOfBlocks){
	__shared__ key temp[256]; // allocated on invocation 
	__shared__ int temp1[16];
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	temp[tx].zOrderValue = g_data[thid].zOrderValue;
	//temp[tx].zOrderValue = g_data[thid].data;
	temp[tx].index=thid;
	if(tx<BYTES){
		temp1[tx]=histogram[tx*numOfBlocks+bx];
	}
	if(tx==0){
		for(int i=1;i<BYTES;i++){
			temp1[i]=temp1[i]+temp1[i-1];
		}
	}
	__syncthreads();
	int index=(temp[tx].zOrderValue>>round)&0x0000000F;
	int pos=position[index*numOfBlocks+bx]-(temp1[index]-tx);
	g_data_out[pos]=g_data[temp[tx].index];
}
__global__ void scan1(int *g_data,int *g_odata,int *output,int length,int size,int flag) {  
	extern __shared__ int temp[]; // allocated on invocation 
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	int pout = 0, pin = 1;
	// load input into shared memory. 
	// This is exclusive scan, so shift right by one and set first elt to 0 
	//temp[pout*n + thid] = (thid > 0) ? g_idata[thid-1] : 0; 
	if(thid<=length-1){ 
		temp[pout*size + tx] = g_data[thid];
		//temp[pin*n + thid] = 0.0;
		__syncthreads();
		for (int offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp[pout*size+tx] = temp[pin*size+tx - offset]+temp[pin*size+tx];
				//temp[pout*n+thid] += temp[pin*n+thid - offset];
			}
			else {
				temp[pout*size+tx] = temp[pin*size+tx];
			}
			__syncthreads(); 
		}
		g_odata[thid] = temp[pout*size+tx]; // write output
		if((tx==size-1) && flag) {
			output[bx]=temp[pout*size+tx];
		}
	}
}
__global__ void scan2(int *g_data, int *output,int length,int size,int numOfBlocks) {
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	if((thid<=length-1)&&(bx!=0)){ 
		g_data[thid]=g_data[thid]+output[bx-1];	
	}
}

//data already in global memory.
void sortGM(datatype *mbrs,int length){

  	unsigned long long int maxZOrderValue=0xffffffffffffffff;
  	bounding_box * mbrs_bak;
    int size=256;
    int numOfBlocks=(length-1)/size+1;
    int i;
    //printf("numOfBlocks=%d\n",numOfBlocks);
    int len=numOfBlocks*size;
	bounding_box * packs=(bounding_box *)malloc((len-length)*sizeof(bounding_box));
	for(i=0;i<(len-length);i++){
		packs[i].zOrderValue=maxZOrderValue;
	}

    int* histogram=0;
    int* position=0;

    (cudaMalloc( (void**) &histogram, numOfBlocks*16*sizeof(int)));
    (cudaMalloc( (void**) &position, numOfBlocks*16*sizeof(int)));
	
	int firstBlocks=(numOfBlocks*BYTES-1)/512+1;
	int* firstOutput=0;
	(cudaMalloc( (void**) &firstOutput, firstBlocks*sizeof(datatype)));
   	(cudaMalloc( (void**) &mbrs_bak, sizeof(bounding_box)*len));
	(cudaMemcpy( (mbrs+length), packs,sizeof(bounding_box)*(len-length),cudaMemcpyHostToDevice));		

	
	dim3 dimBlockSort(size,1,1);
    dim3 dimGridSort(numOfBlocks,1);
	
	dim3 dimBlockScan(512,1,1);
    dim3 dimGridScan(firstBlocks,1);
    dim3 singleGrid(1,1);
	    
	
	for(i=0;i<16;i++){
		LocalSort<<<dimGridSort, dimBlockSort>>>(mbrs,mbrs_bak,histogram,size,i*4,numOfBlocks);
		scan1<<<dimGridScan, dimBlockScan,(2*512*sizeof(int))>>>(histogram,position,firstOutput,numOfBlocks*BYTES,512,1);
		if(firstBlocks>1){
			scan1<<<singleGrid, dimBlockScan,(2*512*sizeof(int))>>>(firstOutput,firstOutput,firstOutput,firstBlocks,512,0);
    		scan2<<<dimGridScan, dimBlockScan>>>(position,firstOutput,len,512,firstBlocks);
		}
		Reorder<<<dimGridSort, dimBlockSort>>>(mbrs_bak,mbrs,histogram,position,size,i*4,numOfBlocks);
	}
	
}

//parallel sorting (odd even sorting)
__global__ void ParallelSorting(bounding_box * pointers,int num)
{
		int i;
		int index=threadIdx.x;
		int bx=blockDim.x;
		bounding_box swap;
		__shared__ int done[2];
		done[0]=0;
		done[1]=0;
		int count=0;
		int oddCycle=0;
		//for(j=0;j<=num;j++){
		do{		
	        //done[0]=1;
	        //done[1]=1;
	        done[oddCycle]=1;
	        //at the end of the array, we add one dummy element to avoid branching.
	        for (i = index*2+oddCycle; i < num; i=i+bx*2) 
    	    {
	 			if(pointers[i].zOrderValue>pointers[i+1].zOrderValue){
	 				swap=pointers[i+1];
	 				pointers[i+1]=pointers[i];
	 				pointers[i]=swap;
	 				done[oddCycle]=0;
	 			}
        	} //for
        	oddCycle = !(oddCycle);
        	count++;
         	__syncthreads();
		//}
		}while((!done[0])||(!done[1]));
}
