//Efficient radix sort implementation on CUDA

// Satish, N., Harris, M., and Garland, M. "Designing Efficient Sorting 
// Algorithms for Manycore GPUs". In Proceedings of IEEE International
// Parallel & Distributed Processing Symposium 2009 (IPDPS 2009).


#include "radixSortModu.h"

__global__ void LocalSort(datatype *g_data,int *histogram,int size,int round,int numOfBlocks);
__global__ void Reorder(datatype *g_data,datatype *g_data_out,int *histogram,int *position,int size,int round,int numOfBlocks);
__global__ void scan1(int *g_data,int *g_odata,int *output,int length,int size,int flag); 
__global__ void scan2(int *g_data, int *output,int length,int size,int numOfBlocks);
__global__ void move(datatype *index,int *rect1_gpu,int *rect2_gpu,int *rect3_gpu,int *rect4_gpu,int *bb1,int *bb2,int *bb3,int *bb4,int base,int num);

__global__ void LocalSort(datatype *g_data,int *histogram,int size,int round,int numOfBlocks){
	extern __shared__ datatype temp[]; //allocated on invocation 
	int tx=threadIdx.x;
	int txx=tx+size+size;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	int i;
	int pout = 0, pin = 1;
	int out=1,in=0;
	int zeros;
	int offset;
	temp[tx] = g_data[thid];
	for(i=0;i<BIT;i++){
		//Count the number of zeros
		temp[pout*size+txx].zOrderValue=(1-((temp[in*size+tx].zOrderValue>>(i+round))&0x00000001));
		__syncthreads();
		for (offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp[pout*size+txx].zOrderValue = temp[pin*size+txx-offset].zOrderValue+temp[pin*size+txx].zOrderValue;
			}
			else {
				temp[pout*size+txx].zOrderValue = temp[pin*size+txx].zOrderValue;
			}
			__syncthreads(); 
		}
		temp[out*size+tx].zOrderValue=(temp[pout*size+txx].zOrderValue)*(1-((temp[in*size+tx].zOrderValue)>>(i+round))&0x00000001);
		zeros=temp[pout*size+3*size-1].zOrderValue;
		//missing
		__syncthreads();
		temp[pout*size+txx].zOrderValue=(((temp[in*size+tx].zOrderValue)>>(i+round))&0x00000001);
		__syncthreads();
		for (offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp[pout*size+txx].zOrderValue = temp[pin*size+txx-offset].zOrderValue+temp[pin*size+txx].zOrderValue;
			}
			else {
				temp[pout*size+txx].zOrderValue = temp[pin*size+txx].zOrderValue;
			}
			__syncthreads(); 
		}
		temp[pin*size+txx].zOrderValue=(temp[pout*size+txx].zOrderValue+zeros)*(((temp[in*size+tx].zOrderValue)>>(i+round))&0x00000001)+temp[out*size+tx].zOrderValue;
		__syncthreads(); 
		temp[out*size+(temp[pin*size+txx].zOrderValue-1)]=temp[in*size+tx];
		out = 1 - out; // swap double buffer indices 
		in = 1 - out;
		__syncthreads(); 
	}
	g_data[thid]=temp[in*size+tx];  
	for(i=0;i<BYTES;i++){
		if((((temp[in*size+tx].zOrderValue)>>(round))&0x0000000F)==i){
			temp[pin*size+txx].zOrderValue=1;
		}
		else{
			temp[pin*size+txx].zOrderValue=0;
		}
		__syncthreads();
		//g_data[thid]=temp[pin*size+txx]; 
		int offset=size;		
		while(offset>1){
			offset=(offset>>1);
			if(tx<offset){
				temp[pin*size+txx].zOrderValue=temp[pin*size+txx+offset].zOrderValue+temp[pin*size+txx].zOrderValue;
			}
			__syncthreads();
		}
		if(tx==0){
			histogram[i*numOfBlocks+bx]=temp[pin*size+txx].zOrderValue;
		}
		__syncthreads(); 
	}
}
__global__ void Reorder(datatype *g_data,datatype *g_data_out,int *histogram,int *position,int size,int round,int numOfBlocks){
	extern __shared__ datatype temp[]; // allocated on invocation 
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	temp[tx] = g_data[thid];
	if(tx<BYTES){
		temp[size+tx].zOrderValue=histogram[tx*numOfBlocks+bx];
	}
	if(tx==0){
		for(int i=1;i<BYTES;i++){
			temp[size+i].zOrderValue=temp[size+i].zOrderValue+temp[size+i-1].zOrderValue;
		}
	}
	__syncthreads();
	int index=((temp[tx].zOrderValue)>>round)&0x0000000F;
	int pos=position[index*numOfBlocks+bx]-(temp[size+index].zOrderValue-tx);
	g_data_out[pos]=temp[tx];
}
__global__ void scan1(int *g_data,int *g_odata,int *output,int length,int size,int flag) {  
	extern __shared__ int temp1[]; // allocated on invocation 
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	int pout = 0, pin = 1;
	// load input into shared memory. 
	// This is exclusive scan, so shift right by one and set first elt to 0 
	//temp[pout*n + thid] = (thid > 0) ? g_idata[thid-1] : 0; 
	if(thid<=length-1){ 
		temp1[pout*size + tx] = g_data[thid];
		//temp[pin*n + thid] = 0.0;
		__syncthreads();
		for (int offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp1[pout*size+tx] = temp1[pin*size+tx - offset]+temp1[pin*size+tx];
				//temp[pout*n+thid] += temp[pin*n+thid - offset];
			}
			else {
				temp1[pout*size+tx] = temp1[pin*size+tx];
			}
			__syncthreads(); 
		}
		g_odata[thid] = temp1[pout*size+tx]; // write output
		if((tx==size-1) && flag) {
			output[bx]=temp1[pout*size+tx];
		}
	}
} 
__global__ void scan2(int *g_data, int *output,int length,int size,int numOfBlocks) {
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	if((thid<=length-1)&&(bx!=0)){ 
		g_data[thid]=g_data[thid]+output[bx-1];	
	}
}
__global__ void move(datatype *zOrder,int *rect1_gpu,int *rect2_gpu,int *rect3_gpu,int *rect4_gpu,int *rect1_rtree,int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int base,int num){
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int thid = tx+bx*blockDim.x;
	if(thid<num){
		rect1_rtree[thid+base]=rect1_gpu[zOrder[thid].index];
		rect2_rtree[thid+base]=rect2_gpu[zOrder[thid].index];
		rect3_rtree[thid+base]=rect3_gpu[zOrder[thid].index];
		rect4_rtree[thid+base]=rect4_gpu[zOrder[thid].index];
		//if((rect1_rtree[thid+base]<0)||(rect2_rtree[thid+base]<0)||(rect3_rtree[thid+base]<0)||(rect4_rtree[thid+base]<0)){
		//	printf("%d--%d--(%d,%d,%d,%d)\n",zOrder[thid].index,thid,rect1_rtree[thid+base],rect3_rtree[thid+base],rect3_rtree[thid+base],rect4_rtree[thid+base]);
		//}
	}	
}

//data already in global memory.
void sortGM(datatype *g_data,int length){
	int i; 
    //int length=1000000;
    int size=128;
    int numOfBlocks=(length-1)/size+1;
    int len=numOfBlocks*size;
    int max=0xffffffff;
    datatype* datapack;
    
    datatype* g_data_out=0;
    int* histogram=0;
    int* position=0;

    (cudaMalloc( (void**) &g_data_out, len*sizeof(datatype)));
    (cudaMalloc( (void**) &histogram, numOfBlocks*16*sizeof(int)));
    (cudaMalloc( (void**) &position, numOfBlocks*16*sizeof(int)));
	
	if(length<len){
	    datapack = (datatype*) malloc((len-length)*sizeof(datatype));
	    for(i = 0; i < (len-length); i++) 
    	{
        	datapack[i].zOrderValue = max;
        	datapack[i].index=max;//(int)(10 * rand()/32768.f);
    	}
	}
	if(length<len){
    	(cudaMemcpy((g_data+length), datapack, sizeof(datatype)*(len-length), cudaMemcpyHostToDevice) );
    }
	int firstBlocks=(numOfBlocks*BYTES-1)/size+1;
	int* firstOutput=0;
	(cudaMalloc( (void**) &firstOutput, firstBlocks*sizeof(datatype)));

    dim3 dimBlock(size,1,1);
    dim3 dimGrid(numOfBlocks,1);
	
	dim3 dimBlockScan(size,1,1);
    dim3 dimGridScan(firstBlocks,1);
    dim3 singleGrid(1,1);
	
	
	/*cudaEvent_t start_event, stop_event;
    cudaEventCreate(&start_event) ;
    cudaEventCreate(&stop_event) ;
    cudaEventRecord(start_event, 0);*/
	for(i=0;i<8;i=i+2){
		LocalSort<<<dimGrid, dimBlock,4*size*sizeof(datatype)>>>(g_data,histogram,size,i*4,numOfBlocks);
		scan1<<<dimGridScan, dimBlockScan,(2*size*sizeof(int))>>>(histogram,position,firstOutput,numOfBlocks*BYTES,size,1);
		scan1<<<singleGrid, dimBlockScan,(2*size*sizeof(int))>>>(firstOutput,firstOutput,firstOutput,firstBlocks,size,0);
    	scan2<<<dimGridScan, dimBlockScan>>>(position,firstOutput,len,size,firstBlocks);
		Reorder<<<dimGrid, dimBlock,((size+BYTES)*sizeof(datatype))>>>(g_data,g_data_out,histogram,position,size,i*4,numOfBlocks);
	
		LocalSort<<<dimGrid, dimBlock,(4*size*sizeof(datatype))>>>(g_data_out,histogram,size,(i+1)*4,numOfBlocks);
		scan1<<<dimGridScan, dimBlockScan,(2*size*sizeof(int))>>>(histogram,position,firstOutput,numOfBlocks*BYTES,size,1);
		scan1<<<singleGrid, dimBlockScan,(2*size*sizeof(int))>>>(firstOutput,firstOutput,firstOutput,firstBlocks,size,0);
    	scan2<<<dimGridScan, dimBlockScan>>>(position,firstOutput,len,size,firstBlocks);
		Reorder<<<dimGrid, dimBlock,((size+BYTES)*sizeof(datatype))>>>(g_data_out,g_data,histogram,position,size,(i+1)*4,numOfBlocks);
	}
	
	cudaThreadSynchronize();
    /*cudaEventRecord(stop_event, 0);
    cudaEventSynchronize(stop_event);
    float time_kernel; 
    cudaEventElapsedTime(&time_kernel, start_event, stop_event);
	printf("Total time %f\n", time_kernel);*/
	//
    //printf("sum=%d\n",sum);
    
    //cudaFree(g_data);
	cudaFree(g_data_out);
	cudaFree(histogram);
	cudaFree(position);
	cudaFree(firstOutput);
}


void moveData(datatype *zOrder,int *rect1_gpu,int *rect2_gpu,int *rect3_gpu,int *rect4_gpu,int *rect1_rtree,int *rect2_rtree,int *rect3_rtree,int *rect4_rtree,int base,int num){
	dim3 dimBlock(512,1);
    dim3 dimGrid((num-1)/512+1,1);
    move<<<dimGrid,dimBlock>>>(zOrder,rect1_gpu,rect2_gpu,rect3_gpu,rect4_gpu,rect1_rtree,rect2_rtree,rect3_rtree,rect4_rtree,base,num);
}
