//Efficient radix sort implementation on CUDA

// Satish, N., Harris, M., and Garland, M. "Designing Efficient Sorting 
// Algorithms for Manycore GPUs". In Proceedings of IEEE International
// Parallel & Distributed Processing Symposium 2009 (IPDPS 2009).


#include "radixSortModu1.h"

__global__ void LocalSort1(rtree_search *g_data,int *histogram,int size,int round,int numOfBlocks);
__global__ void Reorder1(rtree_search *g_data,rtree_search *g_data_out,int *histogram,int *position,int size,int round,int numOfBlocks);
__global__ void scan11(int *g_data,int *g_odata,int *output,int length,int size,int flag); 
__global__ void scan21(int *g_data, int *output,int length,int size,int numOfBlocks);

__global__ void LocalSort1(rtree_search *g_data,int *histogram,int size,int round,int numOfBlocks){
	extern __shared__ rtree_search temp_1[]; //allocated on invocation 
	int tx=threadIdx.x;
	int txx=tx+size+size;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	int i;
	int pout = 0, pin = 1;
	int out=1,in=0;
	int zeros;
	int offset;
	temp_1[tx] = g_data[thid];
	for(i=0;i<BIT;i++){
		//Count the number of zeros
		temp_1[pout*size+txx].nodeid=(1-((temp_1[in*size+tx].nodeid>>(i+round))&0x00000001));
		__syncthreads();
		for (offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp_1[pout*size+txx].nodeid = temp_1[pin*size+txx-offset].nodeid+temp_1[pin*size+txx].nodeid;
			}
			else {
				temp_1[pout*size+txx].nodeid = temp_1[pin*size+txx].nodeid;
			}
			__syncthreads(); 
		}
		temp_1[out*size+tx].nodeid=(temp_1[pout*size+txx].nodeid)*(1-((temp_1[in*size+tx].nodeid)>>(i+round))&0x00000001);
		zeros=temp_1[pout*size+3*size-1].nodeid;
		//missing
		__syncthreads();
		temp_1[pout*size+txx].nodeid=(((temp_1[in*size+tx].nodeid)>>(i+round))&0x00000001);
		__syncthreads();
		for (offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp_1[pout*size+txx].nodeid = temp_1[pin*size+txx-offset].nodeid+temp_1[pin*size+txx].nodeid;
			}
			else {
				temp_1[pout*size+txx].nodeid = temp_1[pin*size+txx].nodeid;
			}
			__syncthreads(); 
		}
		temp_1[pin*size+txx].nodeid=(temp_1[pout*size+txx].nodeid+zeros)*(((temp_1[in*size+tx].nodeid)>>(i+round))&0x00000001)+temp_1[out*size+tx].nodeid;
		__syncthreads(); 
		temp_1[out*size+(temp_1[pin*size+txx].nodeid-1)]=temp_1[in*size+tx];
		out = 1 - out; // swap double buffer indices 
		in = 1 - out;
		__syncthreads(); 
	}
	g_data[thid]=temp_1[in*size+tx];  
	for(i=0;i<BYTES;i++){
		if((((temp_1[in*size+tx].nodeid)>>(round))&0x0000000F)==i){
			temp_1[pin*size+txx].nodeid=1;
		}
		else{
			temp_1[pin*size+txx].nodeid=0;
		}
		__syncthreads();
		//g_data[thid]=temp[pin*size+txx]; 
		int offset=size;		
		while(offset>1){
			offset=(offset>>1);
			if(tx<offset){
				temp_1[pin*size+txx].nodeid=temp_1[pin*size+txx+offset].nodeid+temp_1[pin*size+txx].nodeid;
			}
			__syncthreads();
		}
		if(tx==0){
			histogram[i*numOfBlocks+bx]=temp_1[pin*size+txx].nodeid;
		}
		__syncthreads(); 
	}
}
__global__ void Reorder1(rtree_search *g_data,rtree_search *g_data_out,int *histogram,int *position,int size,int round,int numOfBlocks){
	extern __shared__ rtree_search temp_2[]; // allocated on invocation 
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	temp_2[tx] = g_data[thid];
	if(tx<BYTES){
		temp_2[size+tx].nodeid=histogram[tx*numOfBlocks+bx];
	}
	if(tx==0){
		for(int i=1;i<BYTES;i++){
			temp_2[size+i].nodeid=temp_2[size+i].nodeid+temp_2[size+i-1].nodeid;
		}
	}
	__syncthreads();
	int index=((temp_2[tx].nodeid)>>round)&0x0000000F;
	int pos=position[index*numOfBlocks+bx]-(temp_2[size+index].nodeid-tx);
	g_data_out[pos]=temp_2[tx];
}
__global__ void scan11(int *g_data,int *g_odata,int *output,int length,int size,int flag) {  
	extern __shared__ int temp1[]; // allocated on invocation 
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	int pout = 0, pin = 1;
	// load input into shared memory. 
	// This is exclusive scan, so shift right by one and set first elt to 0 
	//temp[pout*n + thid] = (thid > 0) ? g_idata[thid-1] : 0; 
	if(thid<=length-1){ 
		temp1[pout*size + tx] = g_data[thid];
		//temp[pin*n + thid] = 0.0;
		__syncthreads();
		for (int offset = 1; offset < size; offset *= 2) {
			pout = 1 - pout; // swap double buffer indices 
			pin = 1 - pout;
			if (tx >= offset) {
				temp1[pout*size+tx] = temp1[pin*size+tx - offset]+temp1[pin*size+tx];
				//temp[pout*n+thid] += temp[pin*n+thid - offset];
			}
			else {
				temp1[pout*size+tx] = temp1[pin*size+tx];
			}
			__syncthreads(); 
		}
		g_odata[thid] = temp1[pout*size+tx]; // write output
		if((tx==size-1) && flag) {
			output[bx]=temp1[pout*size+tx];
		}
	}
} 
__global__ void scan21(int *g_data, int *output,int length,int size,int numOfBlocks) {
	int tx=threadIdx.x;
	int bx=blockIdx.x;
	int base=bx*size;
	int thid = tx+base;
	if((thid<=length-1)&&(bx!=0)){ 
		g_data[thid]=g_data[thid]+output[bx-1];	
	}
}


//data already in global memory.
void sortGM1(rtree_search *g_data,int length){
	int i; 
    //int length=1000000;
    //int size=128;
    int size=512;
    int numOfBlocks=(length-1)/size+1;
    int len=numOfBlocks*size;
    int max=0xffffffff;
    rtree_search * datapack;
    
    rtree_search * g_data_out=0;
    int* histogram=0;
    int* position=0;

    (cudaMalloc( (void**) &g_data_out, len*sizeof(rtree_search)));
    (cudaMalloc( (void**) &histogram, numOfBlocks*16*sizeof(int)));
    (cudaMalloc( (void**) &position, numOfBlocks*16*sizeof(int)));
	
	if(length<len){
	    datapack = (rtree_search*) malloc((len-length)*sizeof(rtree_search));
	    for(i = 0; i < (len-length); i++) 
    	{
        	datapack[i].nodeid = max;
        	datapack[i].queryid=max;
    	}
	}
	if(length<len){
    	(cudaMemcpy((g_data+length), datapack, sizeof(rtree_search)*(len-length), cudaMemcpyHostToDevice) );
    }
	int firstBlocks=(numOfBlocks*BYTES-1)/size+1;
	int* firstOutput=0;
	(cudaMalloc( (void**) &firstOutput, firstBlocks*sizeof(rtree_search)));

    dim3 dimBlock(size,1,1);
    dim3 dimGrid(numOfBlocks,1);
	
	dim3 dimBlockScan(size,1,1);
    dim3 dimGridScan(firstBlocks,1);
    dim3 singleGrid(1,1);
	
	
	/*cudaEvent_t start_event, stop_event;
    cudaEventCreate(&start_event) ;
    cudaEventCreate(&stop_event) ;
    cudaEventRecord(start_event, 0);*/
	for(i=0;i<8;i=i+2){
		LocalSort1<<<dimGrid, dimBlock,4*size*sizeof(rtree_search)>>>(g_data,histogram,size,i*4,numOfBlocks);
		scan11<<<dimGridScan, dimBlockScan,(2*size*sizeof(int))>>>(histogram,position,firstOutput,numOfBlocks*BYTES,size,1);
		scan11<<<singleGrid, dimBlockScan,(2*size*sizeof(int))>>>(firstOutput,firstOutput,firstOutput,firstBlocks,size,0);
    	scan21<<<dimGridScan, dimBlockScan>>>(position,firstOutput,len,size,firstBlocks);
		Reorder1<<<dimGrid, dimBlock,((size+BYTES)*sizeof(rtree_search))>>>(g_data,g_data_out,histogram,position,size,i*4,numOfBlocks);
	
		LocalSort1<<<dimGrid, dimBlock,(4*size*sizeof(rtree_search))>>>(g_data_out,histogram,size,(i+1)*4,numOfBlocks);
		scan11<<<dimGridScan, dimBlockScan,(2*size*sizeof(int))>>>(histogram,position,firstOutput,numOfBlocks*BYTES,size,1);
		scan11<<<singleGrid, dimBlockScan,(2*size*sizeof(int))>>>(firstOutput,firstOutput,firstOutput,firstBlocks,size,0);
    	scan21<<<dimGridScan, dimBlockScan>>>(position,firstOutput,len,size,firstBlocks);
		Reorder1<<<dimGrid, dimBlock,((size+BYTES)*sizeof(rtree_search))>>>(g_data_out,g_data,histogram,position,size,(i+1)*4,numOfBlocks);
	}
	
	cudaThreadSynchronize();
    /*cudaEventRecord(stop_event, 0);
    cudaEventSynchronize(stop_event);
    float time_kernel; 
    cudaEventElapsedTime(&time_kernel, start_event, stop_event);
	printf("Total time %f\n", time_kernel);*/
	//
    //printf("sum=%d\n",sum);
    
    //cudaFree(g_data);
	cudaFree(g_data_out);
	cudaFree(histogram);
	cudaFree(position);
	cudaFree(firstOutput);
}
//data in CPU memory.

